base = '..\data\';
fileName = 'input.CR2';
tiffFileName = strcat(strtok(fileName,'.'), '.tiff');
pythonScript = '..\predict\predict_dual_iso_frames.py';
dualIsoMatName = 'dual_iso.mat';
highMatName = 'high.mat';
lowMatName = 'low.mat';
highModelName = '../models/model_high.h5';
lowModelName = '../models/model_low.h5';
isSimulation = 1;
convertToTiffCommand = '..\dcraw\dcraw.exe -4 -D -T ';
getParametersCommand = '..\dcraw\dcraw.exe -v -w -T ';
bayer_pattern = 'rggb';
high_exp = 16;
low_exp = 1;

[~,cmdout] = system([getParametersCommand base fileName]);
black = str2double(regexp(cmdout, '(?<=Scaling with darkness[^0-9]*)[0-9]*\.?[0-9]+', 'match'));
saturation = str2double(regexp(cmdout, '(?<=saturation[^0-9]*)[0-9]*\.?[0-9]+', 'match'));
test = cmdout(strfind(cmdout, 'multipliers'):strfind(cmdout, 'AHD'));
wb_multipliers = sscanf(test, '%*s %f %f %f %f');

[~, cmdout] = system([convertToTiffCommand base fileName]);

%Copied from dcraw%
%EOS 600 D% 
xyz2cam = [ 6461 -907 -882 
    -4300 12184 2378 
    -819 1944 5931] / 10000; 

% EOS 5D MARK III
% xyz2cam = [ 6722 -635 -963
%     -4287 12460 2028 
%     -908 2162 5668 ] / 12460;

srgb2xyz = [0.4124564 0.3575761 0.1804375;
    0.2126729 0.7151522 0.0721750;
    0.0193339 0.1191920 0.9503041];

img = imread([base tiffFileName]);
% img = img(1:1024, 1:1024);

[rows, columns, numberOfColorChannels] = size(img);
imagePath = strcat(base, strtok(fileName,'.'));

tic
img = (double(img)-black)/(saturation-black);
img = max(0,min(img,1));

save(dualIsoMatName, 'img');
[status, cmdout] = system(['py ' pythonScript ' ' dualIsoMatName ' ' highModelName ' ' lowModelName ' ' int2str(size(img,1) / 3)]);

high_mat = load(highMatName);
low_mat = load(lowMatName);

imgh_demosaiced = high_mat.high;
imgl_demosaiced = low_mat.low;

mkdir(imagePath);

rgb2cam = xyz2cam * srgb2xyz;
rgb2cam = rgb2cam ./ repmat(sum(rgb2cam,2),1,3);
cam2rgb = rgb2cam^-1;

mask = wbmask(size(imgl_demosaiced,1),size(imgl_demosaiced,2),...   
                            wb_multipliers,bayer_pattern);
imgl_demosaiced = imgl_demosaiced .* mask;
imgh_demosaiced = imgh_demosaiced .* mask;

imgh_demosaiced = demosaic(uint16(imgh_demosaiced * 65535), bayer_pattern);
imgl_demosaiced = demosaic(uint16(imgl_demosaiced * 65535), bayer_pattern);

imgh_demosaiced = double(imgh_demosaiced) / 65535;
imgl_demosaiced = double(imgl_demosaiced) / 65535;

imgh_demosaiced = apply_cmatrix(imgh_demosaiced, cam2rgb);
imgh_demosaiced = max(0,min(imgh_demosaiced,1));

imgl_demosaiced = apply_cmatrix(imgl_demosaiced, cam2rgb);
imgl_demosaiced = max(0,min(imgl_demosaiced,1));

norm_value = 1;

[rows, columns, numOfColor] = size(imgh_demosaiced);
bToSingle = 0;
bNormalization = 1;
stack = zeros(rows, columns, numOfColor, 2, 'single');
for i=1:2
    %read an image, and convert it into floating-point
    if i == 1
        img_tmp = imgh_demosaiced;
    else
        img_tmp = imgl_demosaiced; 
    end
    
    %store in the stack
    if(bToSingle)
        stack(:,:,:,i) = single(img_tmp);   
    else
        stack(:,:,:,i) = img_tmp;   
    end
end
if(bNormalization)
    stack = stack / norm_value;
end

stack_exposure(1) = high_exp;
stack_exposure(2) = low_exp;

hdr = buildHdrDemosaiced(stack, stack_exposure, 'linear', 'Deb97_p05', 'linear');
hdr(isnan(hdr))=0;
toc

exrwrite(hdr, strcat(imagePath, '\hdr_averaged.exr'));
delete *.mat;
delete(strcat(base, tiffFileName));