import os
import random
import re

import cv2
import h5py
import keras.backend as K
import numpy as np
from keras.callbacks import ModelCheckpoint
from keras.layers import Conv2D
from keras.layers import Input, Activation, Add
from keras.models import Model
from keras.optimizers import adam
from keras.preprocessing.image import ImageDataGenerator

model_name = "model_high"
dataset_name = "rain_dataset_additive_high_136.h5"

DATA_PATH = "dataset/"

noiseValuesFileName = 'noiseRatios.txt'
noiseValues = {}
patch_size = 136
label_size = 136

epoch = 60
batch_size = 64

validation_split = 0.01

black = 0
saturation = 4095

mu = 0
sigmaEpsilon = 0
startSigma = 0
sigma = 5

countLimit = 4
augmentationEnabled = True

patch_repeating = 1

samplingRate = 10

lowNoiseDivider = 4.0

dummySize = 200000
num_filters = 32
num_res_blocks = 10
res_block_scaling = None
conv_counter = 0
widthFull = 5168
heightFull = 3400

dilation_array = [1, 1, 1, 2, 2, 2, 2, 4, 4, 4, 4, 8, 8, 4, 4, 4, 4, 2, 2, 2, 2, 1, 1]

def model_layer(Adam):
    input = Input(shape=(None, None, 1))
    global conv_counter
    conv_counter = 0
    conv1 = Conv2D(num_filters, 3, padding='same', activaion="relu", kernel_initializer="glorot_uniform",
                   use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(input)
    conv_counter = conv_counter + 1

    conv2 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv1)
    conv_counter = conv_counter + 1
    conv2_activation = Activation('relu')(conv2)
    conv3 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv2_activation)
    conv_counter = conv_counter + 1
    conv3_merged = Add()([conv1, conv3])

    conv4 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv3_merged)
    conv_counter = conv_counter + 1
    conv4_activation = Activation('relu')(conv4)
    conv5 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv4_activation)
    conv_counter = conv_counter + 1
    conv5_merged = Add()([conv3_merged, conv5])

    conv6 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv5_merged)
    conv_counter = conv_counter + 1
    conv6_activation = Activation('relu')(conv6)
    conv7 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv6_activation)
    conv_counter = conv_counter + 1
    conv7_merged = Add()([conv5_merged, conv7])

    conv8 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv7_merged)
    conv_counter = conv_counter + 1
    conv8_activation = Activation('relu')(conv8)
    conv9 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                   dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv8_activation)
    conv_counter = conv_counter + 1
    conv9_merged = Add()([conv7_merged, conv9])

    conv10 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv9_merged)
    conv_counter = conv_counter + 1
    conv10_activation = Activation('relu')(conv10)
    conv11 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv10_activation)
    conv_counter = conv_counter + 1
    conv11_merged = Add()([conv9_merged, conv11])

    conv12 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv11_merged)
    conv_counter = conv_counter + 1
    conv12_activation = Activation('relu')(conv12)
    conv13 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv12_activation)
    conv_counter = conv_counter + 1
    conv13_merged = Add()([conv11_merged, conv13])
    conv13_merged = Add()([conv13_merged, conv9_merged])

    conv14 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv13_merged)
    conv_counter = conv_counter + 1
    conv14_activation = Activation('relu')(conv14)
    conv15 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv14_activation)
    conv_counter = conv_counter + 1
    conv15_merged = Add()([conv13_merged, conv15])
    conv15_merged = Add()([conv15_merged, conv7_merged])

    conv16 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv15_merged)
    conv_counter = conv_counter + 1
    conv16_activation = Activation('relu')(conv16)
    conv17 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv16_activation)
    conv_counter = conv_counter + 1
    conv17_merged = Add()([conv15_merged, conv17])
    conv17_merged = Add()([conv17_merged, conv5_merged])

    conv18 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv17_merged)
    conv_counter = conv_counter + 1
    conv18_activation = Activation('relu')(conv18)
    conv19 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv18_activation)
    conv_counter = conv_counter + 1
    conv19_merged = Add()([conv17_merged, conv19])
    conv19_merged = Add()([conv19_merged, conv3_merged])

    conv20 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv19_merged)
    conv_counter = conv_counter + 1
    conv20_activation = Activation('relu')(conv20)
    conv21 = Conv2D(num_filters, 3, padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv20_activation)
    conv_counter = conv_counter + 1
    conv21_merged = Add()([conv19_merged, conv21])
    conv21_merged = Add()([conv21_merged, conv1])

    conv22 = Conv2D(num_filters, 3, activation='relu', padding='same', kernel_initializer="glorot_uniform",
                    use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(
        conv21_merged)
    conv_counter = conv_counter + 1
    conv23 = Conv2D(1, 3, activation='relu', padding='same', kernel_initializer="glorot_uniform", use_bias="True",
                    dilation_rate=(dilation_array[conv_counter], dilation_array[conv_counter]))(conv22)
    conv_counter = conv_counter + 1
    model = Model(input, conv23, name="high")
    model.compile(optimizer=Adam, loss=loss, metrics=['mean_squared_error'])
    return model


def loss(y_true, y_pred):
    loss = K.mean(K.square(y_true - y_pred))
    return loss


def train():
    data, label = read_training_data(dataset_name)
    initial_lr = 0.0001
    Adam = adam(lr=initial_lr)
    model = model_layer(Adam)

    filepath = model_name + "-weights-{epoch:02d}-{loss:.2f}.h5"
    checkpoint = ModelCheckpoint(filepath, monitor='loss', verbose=1,
                                 save_best_only=True, save_weights_only=True,
                                 mode='min', period=1)

    callbacks_list = [checkpoint]

    model.fit(data, label, validation_split=validation_split, batch_size=batch_size, epochs=epoch, verbose=1,
              callbacks=callbacks_list)

    model.save_weights(model_name)


def read_training_data(file):
    with h5py.File(file, 'r') as hf:
        data = np.array(hf.get('data'))
        label = np.array(hf.get('label'))
        train_data = np.transpose(data, (0, 2, 3, 1))
        train_label = np.transpose(label, (0, 2, 3, 1))
        return train_data, train_label


def write_hdf5(data, labels, output_filename):
    x = data.astype(np.float32)
    y = labels.astype(np.float32)

    with h5py.File(output_filename, 'w') as h:
        h.create_dataset('data', data=x, shape=x.shape, maxshape=(None, None, None, None), chunks=True)
        h.create_dataset('label', data=y, shape=y.shape, maxshape=(None, None, None, None), chunks=True)


def write_hdf5_additional(data, labels, output_filename):
    x = data.astype(np.float32)
    y = labels.astype(np.float32)

    with h5py.File(output_filename, 'a') as h:
        h["data"].resize((h["data"].shape[0] + x.shape[0]), axis=0)
        h["data"][-x.shape[0]:] = x

        h["label"].resize((h["label"].shape[0] + y.shape[0]), axis=0)
        h["label"][-y.shape[0]:] = y


def createTrainDataset():
    names_dir = os.listdir(DATA_PATH)
    nums_dir = names_dir.__len__()

    for dir in range(nums_dir):
        print(dir)
        dir_path = DATA_PATH + names_dir[dir] + "/"
        names = os.listdir(dir_path)
        nums = names.__len__()

        for i in range(nums):
            patch_number = 0

            data = np.zeros(shape=(dummySize, 1, patch_size, patch_size), dtype=np.double)
            label = np.zeros(shape=(dummySize, 1, patch_size, patch_size), dtype=np.double)

            high = cv2.imread(dir_path + names[i] + "/high.tiff", -1)
            low = cv2.imread(dir_path + names[i] + "/low.tiff", -1)

            high = high[0:heightFull, 0:widthFull]
            low = low[0:heightFull, 0:widthFull]

            highNormalized = high.astype(float) / 65535.
            lowNormalized = low.astype(float) / 65535.

            height, width = highNormalized.shape[:2]

            for y in range(0, height, patch_size):
                for x in range(0, width, patch_size):
                    patch_high = highNormalized[y:y + patch_size, x:x + patch_size]
                    patch_low = lowNormalized[y:y + patch_size, x:x + patch_size]

                    highMean = cv2.mean(patch_high)[0] * 65535
                    highMeanNormalized = highMean / 65535.
                    lowNoiseDivider = noiseValues[str(format(round(highMeanNormalized, 3), '.3f'))]

                    randNumbers = random.sample(range(startSigma * samplingRate, sigma * samplingRate),
                                                patch_repeating)

                    for a in range(patch_repeating):

                        noisyInterlaced = np.zeros((patch_size, patch_size), dtype=np.float)

                        if randNumbers[a] == 0:
                            randNumbers[a] = randNumbers[a] + sigmaEpsilon
                        else:
                            randNumbers[a] = randNumbers[a] / float(samplingRate)

                        gaussHigh = np.random.normal(mu, randNumbers[a] / 255.0, (patch_size, patch_size))
                        gaussHigh = gaussHigh.reshape(patch_size, patch_size)

                        noisyHighPatch = patch_high + gaussHigh

                        gaussLow = np.random.normal(mu, (randNumbers[a] / lowNoiseDivider) / 255.0,
                                                    (patch_size, patch_size))
                        gaussLow = gaussLow.reshape(patch_size, patch_size)

                        noisyLowPatch = patch_low + gaussLow

                        for patch_index in range(patch_size):
                            if patch_index % 4 == 0 or patch_index % 4 == 1:
                                noisyInterlaced[patch_index] = noisyHighPatch[patch_index]
                            else:
                                noisyInterlaced[patch_index] = noisyLowPatch[patch_index]

                        noisyInterlaced = noisyInterlaced.clip(min=0)
                        noisyInterlaced = noisyInterlaced.clip(max=1)

                        data[patch_number, 0, :, :] = noisyInterlaced
                        label[patch_number, 0, :, :] = patch_high

                        patch_number = patch_number + 1

            print(str(patch_number) + " patches have been written into " + dataset_name)
            dummyElementSize = dummySize - patch_number
            data = data[:-dummyElementSize]
            label = label[:-dummyElementSize]

            if dir == 0 and i == 0:
                write_hdf5(data, label, dataset_name)
            else:
                write_hdf5_additional(data, label, dataset_name)

            data = np.zeros((1, 1, 64, 64))
            label = np.zeros((1, 1, 64, 64))
            if augmentationEnabled == True:
                dataGenenerator(dir_path + names[i])


def rgbtoBayer(img):
    height, width = img.shape[:2]
    mosaicedImage = np.zeros((height, width), dtype=np.double)
    for i in range(height):
        for j in range(width):
            if i % 2 == 1 and j % 2 == 1:
                mosaicedImage[i][j] = img[i][j][0]
            elif i % 2 == 1 and j % 2 == 0:
                mosaicedImage[i][j] = img[i][j][1]
            elif i % 2 == 0 and j % 2 == 1:
                mosaicedImage[i][j] = img[i][j][1]
            elif i % 2 == 0 and j % 2 == 0:
                mosaicedImage[i][j] = img[i][j][2]
    return mosaicedImage


def autocrop(image, threshold=0):
    if len(image.shape) == 3:
        flatImage = np.max(image, 2)
    else:
        flatImage = image
    assert len(flatImage.shape) == 2

    rows = np.where(np.max(flatImage, 0) > threshold)[0]
    if rows.size:
        cols = np.where(np.max(flatImage, 1) > threshold)[0]
        image = image[cols[0]: cols[-1] + 1, rows[0]: rows[-1] + 1]
    else:
        image = image[:1, :1]

    return image


def dataGenenerator(imagePath):
    datagen = ImageDataGenerator(
        rotation_range=60,
        width_shift_range=0.2,
        height_shift_range=0.2,
        shear_range=0.2,
        zoom_range=0.2,
        horizontal_flip=True,
        fill_mode='constant'
    )

    low = cv2.imread(imagePath + '/low.tiff', -1)  # this is a PIL image
    high = cv2.imread(imagePath + '/high.tiff', -1)

    high = high[0:heightFull, 0:widthFull]
    low = low[0:heightFull, 0:widthFull]

    colorLow = cv2.cvtColor(low, cv2.COLOR_BAYER_GB2RGB)
    colorHigh = cv2.cvtColor(high, cv2.COLOR_BAYER_GB2RGB)

    height, width = low.shape[:2]
    colorLow = np.asarray(colorLow).reshape(height, width, 3)
    colorLow = colorLow.reshape((1,) + colorLow.shape)

    colorHigh = np.asarray(colorHigh).reshape(height, width, 3)
    colorHigh = colorHigh.reshape((1,) + colorHigh.shape)

    images_flow_high = datagen.flow(colorHigh, batch_size=1, seed=1337)
    images_flow_low = datagen.flow(colorLow, batch_size=1, seed=1337)

    data = np.zeros(shape=(dummySize, 1, patch_size, patch_size), dtype=np.double)
    label = np.zeros(shape=(dummySize, 1, patch_size, patch_size), dtype=np.double)

    patch_number = 0

    for i in range(countLimit):
        high_image = images_flow_high.next().reshape(height, width, 3)
        high_image = rgbtoBayer(high_image)
        high_image = np.uint16(high_image)

        low_image = images_flow_low.next().reshape(height, width, 3)
        low_image = rgbtoBayer(low_image)
        low_image = np.uint16(low_image)

        highNormalized = high_image.astype(float) / 65535.
        lowNormalized = low_image.astype(float) / 65535.

        height, width = low_image.shape[:2]

        for y in range(0, height, patch_size):
            for x in range(0, width, patch_size):
                patch_high = highNormalized[y:y + patch_size, x:x + patch_size]
                patch_low = lowNormalized[y:y + patch_size, x:x + patch_size]

                highMean = cv2.mean(patch_high)[0] * 65535  # finds the mean of the high patch
                highIndices = np.argwhere(patch_high == 0)
                lowIndices = np.argwhere(patch_low == 0)

                if any(elem in highIndices for elem in lowIndices) == False:
                    highMeanNormalized = highMean / 65535.
                    lowNoiseDivider = noiseValues[str(format(round(highMeanNormalized, 3), '.3f'))]

                    randNumbers = random.sample(range(startSigma * samplingRate, sigma * samplingRate),
                                                patch_repeating)

                    for a in range(patch_repeating):
                        noisyInterlaced = np.zeros((patch_size, patch_size), dtype=np.float)

                        if randNumbers[a] == 0:
                            randNumbers[a] = randNumbers[a] + sigmaEpsilon
                        else:
                            randNumbers[a] = randNumbers[a] / float(samplingRate)

                        gaussHigh = np.random.normal(mu, randNumbers[a] / 255.0, (patch_size, patch_size))
                        gaussHigh = gaussHigh.reshape(patch_size, patch_size)

                        noisyHighPatch = patch_high + gaussHigh

                        gaussLow = np.random.normal(mu, (randNumbers[a] / lowNoiseDivider) / 255.0,
                                                    (patch_size, patch_size))
                        gaussLow = gaussLow.reshape(patch_size, patch_size)

                        noisyLowPatch = patch_low + gaussLow

                        for patch_index in range(patch_size):
                            if patch_index % 4 == 0 or patch_index % 4 == 1:
                                noisyInterlaced[patch_index] = noisyHighPatch[patch_index]
                            else:
                                noisyInterlaced[patch_index] = noisyLowPatch[patch_index]

                        noisyInterlaced = noisyInterlaced.clip(min=0)
                        noisyInterlaced = noisyInterlaced.clip(max=1)

                        data[patch_number, 0, :, :] = noisyInterlaced
                        label[patch_number, 0, :, :] = patch_high

                        patch_number = patch_number + 1

    dummyElementSize = dummySize - patch_number
    data = data[:-dummyElementSize]
    label = label[:-dummyElementSize]
    print(str(patch_number) + " augmented patches have been written into " + dataset_name)
    write_hdf5_additional(data, label, dataset_name)


def readNoiseValues():
    file = open(noiseValuesFileName, 'r')
    file.readline()
    for line in file:
        numbers = re.findall(r"[-+]?\d*\.\d+|\d+", line)
        noiseValues[numbers[0]] = float(numbers[1])


if __name__ == "__main__":
    readNoiseValues()
    createTrainDataset()
    train()
